import * as admin from 'firebase-admin';
import Firestore = FirebaseFirestore.Firestore;
import CollectionReference = FirebaseFirestore.CollectionReference
import DocumentReference = FirebaseFirestore.DocumentReference;
import DocumentSnapshot = FirebaseFirestore.DocumentSnapshot;
import Transaction = FirebaseFirestore.Transaction;
import WriteResult = FirebaseFirestore.WriteResult;
import chunk from "lodash.chunk";
import WhereFilterOp = FirebaseFirestore.WhereFilterOp;
import Query = FirebaseFirestore.Query;

const MAX_BATCH_SIZE = 500
export class FirestoreUtil {
    public static decode<T>(documentSnapshot : DocumentSnapshot<T>) : T {
        return  {
            ...this.timestampsToDates(documentSnapshot.data()!),
            created: documentSnapshot.createTime!.toDate(),
            lastModified: documentSnapshot.updateTime!.toDate(),
            id : documentSnapshot.id
        } as T
    }

    public static timestampsToDates<T>(data: T) : T {
        Object.keys(data).forEach(key => {
            if (data[key] && typeof data[key] === 'object' && (data[key]["_seconds"] !== undefined && data[key]["_nanoseconds"] !== undefined))
                data[key] = data[key].toDate();
            else if (data[key] && typeof data[key] === 'object') {
                this.timestampsToDates(data[key])
            }
        });
        return data;
    }

    public static datesToTimestamps<T>(data: T) : T {
        Object.keys(data).forEach(key => {
            if (data[key] && typeof data[key] === 'object' && data[key] instanceof Date) {
                data[key] = admin.firestore.Timestamp.fromDate(data[key])
            } else if (data[key] && typeof data[key] === 'object') {
                data[key] = this.datesToTimestamps(data[key])
            } else if (data[key] && typeof data[key] === 'string') {
                const parsedDate = this.parseIsoDate(data[key])
                if (parsedDate != undefined) {
                    data[key] = admin.firestore.Timestamp.fromDate(parsedDate)
                }
            }
        });
        return data;
    }

    public static parseIsoDate(input: string): Date|undefined {
        if (!/\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}.\d{3}Z/.test(input) &&
            !/\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z/.test(input))
            return undefined;

        return new Date(input)
    }
}

const MAX_WHERE_IN = 10

export class FirestoreClient {
    private readonly firestore: Firestore

    public constructor(firestore: Firestore) {
        this.firestore = firestore
    }

    public getCollection<T>(collectionName: string, optionalFields: string[] = [], versioned = true) : FirestoreCollection<T> {
        return new FirestoreCollection<T>(
            this.firestore.collection(collectionName) as CollectionReference<T>,
            this.firestore,
            optionalFields,
            versioned
        )
    }

    public runTransaction<T>(updateFunction: (transaction: Transaction) => Promise<T>,transactionOptions?:{maxAttempts?: number}
    ): Promise<T> {
        return this.firestore.runTransaction<T>(updateFunction, transactionOptions)
    }
}

export class FirestoreCollection<T> {

    private readonly collectionReference: CollectionReference<T>;
    private readonly firestore : Firestore
    private readonly optionalFields: string[]
    private readonly versioned: boolean

    public constructor(collectionReference: CollectionReference<T>, firestore : Firestore, optionalFields: string[], versioned = true ) {
        this.collectionReference = collectionReference
        this.firestore = firestore
        this.optionalFields = optionalFields
        this.versioned = versioned
    }

    public async documentExists(id: string|DocumentReference<T>, transaction?: Transaction) : Promise<boolean> {
        const docRef = typeof id === "string" ? this.collectionReference.doc(id) : id;
        const documentSnapshot = transaction ? await transaction.get(docRef) : await docRef.get()
        if (documentSnapshot === undefined || !documentSnapshot.exists) {
            return Promise.resolve(false)
        }
        return Promise.resolve(true)
    }

    public async getDocument(id: string|DocumentReference<T>, transaction?: Transaction) : Promise<T | undefined> {
        const docRef = typeof id === "string" ? this.collectionReference.doc(id) : id;
        const documentSnapshot = transaction ? await transaction.get(docRef) : await docRef.get()
        if (documentSnapshot === undefined || !documentSnapshot.exists) {
            return Promise.resolve(undefined)
        }
        return Promise.resolve(FirestoreUtil.decode(documentSnapshot))
    }


    public getSubCollection<V>(documentId: string, collectionName: string, optionalFields: string[] = [], versioned =  true) : FirestoreCollection<V> {
        this.collectionReference.doc(documentId).collection(collectionName)
        return new FirestoreCollection<V>(
            this.collectionReference.doc(documentId).collection(collectionName) as CollectionReference<V>,
            this.firestore,
            optionalFields,
            versioned
        )
    }

    public async setDocument(id: string, data: T) : Promise<WriteResult> {
        this.deleteOptionalFields(data)
        if (this.versioned) this.applyVersionIncrement(FirestoreUtil.datesToTimestamps(data))
        const docRef = this.collectionReference.doc(id);
        return docRef.set(data, {merge : true})
    }

    public async listDocuments() : Promise<T[]> {
        const docRefs = await this.collectionReference.listDocuments();
        if (docRefs !== undefined) {
            return Promise.all(docRefs.map(async doc => FirestoreUtil.decode(await doc.get())))
        }
        return Promise.reject()
    }

    public async addDocument(data: T) : Promise<T | undefined> {
        if (this.versioned) this.applyVersionIncrement(data)
        const docRef : DocumentReference<T> = await this.collectionReference.add(FirestoreUtil.datesToTimestamps(data));
        if (docRef === undefined) {
            return Promise.reject()
        }

        return this.getDocument(docRef)
    }

    public async updateDocument(id: string, updateData: any, transaction?: Transaction) : Promise<T | undefined> {
        if (this.versioned) this.applyVersionIncrement(updateData)
        const docRef = this.collectionReference.doc(id);
        transaction ? await transaction.update(docRef, FirestoreUtil.datesToTimestamps(updateData)) : await docRef.update(updateData)
        return this.getDocument(id)
    }

    public async deleteDocument(id: string) : Promise<WriteResult> {
        return this.collectionReference.doc(id).delete()
    }


    public setDocumentInTransaction(id: string|DocumentReference<T>|undefined, data: T, transaction: Transaction) : Transaction {
        this.deleteOptionalFields(data)
        if (this.versioned) this.applyVersionIncrement(data)
        let docRef
        if (!id) {
            docRef = this.collectionReference.doc()
        } else {
            docRef = typeof id === "string" ? this.collectionReference.doc(id) : id;
        }
        return transaction.set(docRef, FirestoreUtil.datesToTimestamps(data), {merge : true})
    }

    public async executeQuery(fieldPath: string, opStr: WhereFilterOp, value: any) : Promise<T[]> {
        const querySnapshot = await this.collectionReference.where(fieldPath, opStr, value).get()
        if (querySnapshot !== undefined) {
            return querySnapshot.docs.map(doc => FirestoreUtil.decode(doc))
        }
        return Promise.reject()
    }

    public startQuery(fieldPath: string, opStr: WhereFilterOp, value: any) : Query<T> {
        return this.collectionReference.where(fieldPath, opStr, value)
    }

    public async getQueryResult(query: Query<T>) : Promise<T[]> {
        const querySnapshot = await query.get()
        if (querySnapshot !== undefined) {
            return querySnapshot.docs.map(doc => FirestoreUtil.decode(doc))
        }
        return Promise.reject()
    }

    public async bulkSet(data: T[]): Promise<void> {
        await Promise.all(chunk(data, MAX_BATCH_SIZE).map(async dataRefs => {
             const writeBatch = this.firestore.batch();
            (dataRefs).map(value => {
                    this.deleteOptionalFields(value)
                    if (this.versioned) this.applyVersionIncrement(value)
                    const docRef = value.id ? this.collectionReference.doc(value.id) : this.collectionReference.doc()
                    writeBatch.set(
                        docRef,
                        FirestoreUtil.datesToTimestamps(value),
                        {merge: true}
                    )
                }
            )
            await writeBatch.commit()
        })).catch(reason => {
            return Promise.reject()
        })
    }


    public async bulkModification(data: BulkOperation<T>[]): Promise<void> {
        await Promise.all(chunk(data, MAX_BATCH_SIZE).map(async bulkOperations => {
            const writeBatch = this.firestore.batch();
            (bulkOperations).map(bulkOperation => {
                    const docRef = bulkOperation.payloadId ? this.collectionReference.doc(bulkOperation.payloadId) : this.collectionReference.doc()
                    switch (bulkOperation.operation) {
                        case BulkOperationEnum.CREATE:
                        case BulkOperationEnum.UPDATE:
                            this.deleteOptionalFields(bulkOperation.payload)
                            if (this.versioned) this.applyVersionIncrement(bulkOperation.payload)
                            writeBatch.set(
                                docRef,
                                FirestoreUtil.datesToTimestamps(bulkOperation.payload),
                                {merge: true}
                            )
                            break;
                        case BulkOperationEnum.DELETE:
                            writeBatch.delete(this.collectionReference.doc(bulkOperation.payloadId))
                            break;
                        default:
                            break;

                    }
                }
            )
            await writeBatch.commit()
        })).catch(reason => {
            return Promise.reject()
        })
    }

    public async getDocumentsById(ids: string[], transaction?: Transaction) : Promise<T[]> {
        let result : T[] = []
        await Promise.all(chunk(ids, MAX_WHERE_IN).map(async idsChunk => {
            const query = this.collectionReference.where(admin.firestore.FieldPath.documentId(), "in", idsChunk as string[])
            const queryResult = await (transaction ? transaction.get(query) : query.get())
            result = result.concat(queryResult.docs.map(value => FirestoreUtil.decode(value)))
        }))

        return result
    }

    private applyVersionIncrement(data : T) {
        (data as any).version = admin.firestore.FieldValue.increment(1)
    }

    private deleteOptionalFields(data : T) {
        this.optionalFields
            .filter(prop => data[prop] === undefined)
            .map(value => data[value] = admin.firestore.FieldValue.delete())
    }
}

export interface BulkOperation<T> {
    operation: BulkOperationEnum,

    payload?: T,

    payloadId?: string

}

export enum BulkOperationEnum {
    CREATE = <any> 'CREATE',
    UPDATE = <any> 'UPDATE',
    DELETE = <any> 'DELETE'
}
